@@ -1,6 +1,7 @@
 /*
 
-Copyright (C) Jarkko Hietaniemi, 1998,1999,2000. All Rights Reserved.
+Copyright (C) by Jarkko Hietaniemi, 1998,1999,2000,2001,2002,2003,2006.
+All Rights Reserved.
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of either:
@@ -35,6 +36,7 @@ Furthermore:
 #include <string.h>
 #include <stdlib.h>
 #include <ctype.h>
+#include <assert.h>
 
 #define APSE_BITS_IN_BITVEC	(8*sizeof(apse_vec_t))
 
@@ -47,10 +49,10 @@ Furthermore:
 #endif
 
 #define APSE_BIT(i)		((apse_vec_t)1 << ((i)%APSE_BITS_IN_BITVEC))
-#define APSE_IDX(p, q, i)	((p)*(q)+(i)/APSE_BITS_IN_BITVEC)
-#define APSE_BIT_SET(bv, p, q, i) (bv[APSE_IDX(p, q, i)] |=  APSE_BIT(i))
-#define APSE_BIT_CLR(bv, p, q, i) (bv[APSE_IDX(p, q, i)] &= ~APSE_BIT(i))
-#define APSE_BIT_TST(bv, p, q, i) (bv[APSE_IDX(p, q, i)] &   APSE_BIT(i))
+#define APSE_IDX(p, q, i)	((p)*(q)+((i)/APSE_BITS_IN_BITVEC))
+#define APSE_BIT_SET(bv, p, q, i) ((bv[APSE_IDX(p, q, i)] |=  APSE_BIT(i)))
+#define APSE_BIT_CLR(bv, p, q, i) ((bv[APSE_IDX(p, q, i)] &= ~APSE_BIT(i)))
+#define APSE_BIT_TST(bv, p, q, i) ((bv[APSE_IDX(p, q, i)] &   APSE_BIT(i)))
 
 #define APSE_MATCH_STATE_BOT		0
 #define APSE_MATCH_STATE_SEARCH		1
@@ -175,18 +177,21 @@ apse_bool_t apse_set_pattern(apse_t*		ap,
 
     ap->is_greedy    = 0;
 
-    ap->prev_active   = 0;
+    ap->prev_equal   = 0;
+    ap->prev_active  = 0;
 
     ap->pattern_size = pattern_size;
     ap->bitvectors_in_state = (pattern_size - 1)/APSE_BITS_IN_BITVEC + 1;
 
     if (ap->edit_distance)
 	ap->largest_distance = ap->edit_distance * ap->bitvectors_in_state;
+    else
+	ap->largest_distance =  0;
 
     ap->bytes_in_state = ap->bitvectors_in_state * sizeof(apse_vec_t);
 
     ap->case_mask = calloc((apse_size_t)APSE_CHAR_MAX, ap->bytes_in_state);
-    if (!ap->case_mask)
+    if (ap->case_mask == 0)
 	goto out;
 
     for (i = 0; i < pattern_size; i++)
@@ -270,7 +275,7 @@ static int _apse_wrap_slice(apse_t*		ap,
 			    apse_ssize_t*	begin_out,
 			    apse_ssize_t*	size_out) {
     if (begin_in < 0) {
-	if (-begin_in > ap->pattern_size)
+	if ((apse_size_t)-begin_in > ap->pattern_size)
 	    return 0;
 	begin_in = ap->pattern_size + begin_in;
     }
@@ -282,10 +287,10 @@ static int _apse_wrap_slice(apse_t*		ap,
 	begin_in -=  size_in;
     }
 
-    if (begin_in >= ap->pattern_size)
+    if ((apse_size_t)begin_in >= ap->pattern_size)
 	return 0;
 	
-    if (begin_in + size_in > ap->pattern_size)
+    if ((apse_size_t)begin_in + size_in > ap->pattern_size)
 	size_in = ap->pattern_size - begin_in;
 
     if (begin_out)
@@ -327,8 +332,9 @@ apse_bool_t apse_set_charset(apse_t*		ap,
 			     apse_size_t	set_size,
 			     apse_bool_t	complement) {
     apse_size_t	bitvectors_in_state = ap->bitvectors_in_state;
-    apse_ssize_t	true_index, i;
+    apse_ssize_t	true_index;
     apse_bool_t	okay = 0;
+    apse_size_t i;
 
     if (!_apse_wrap_slice(ap, pattern_index, (apse_ssize_t)1,
 			      &true_index, 0))
@@ -362,11 +368,18 @@ static void _apse_reset_state(apse_t* ap) {
     (void)memset(ap->state,      0, ap->bytes_in_all_states);
     (void)memset(ap->prev_state, 0, ap->bytes_in_all_states);
 
+    ap->prev_equal  = 0;
     ap->prev_active = 0;
 
     for (i = 1; i <= ap->edit_distance; i++) {
-	for (j = 0; j < i; j++)
+        for (j = 0; j < i; j++) {
+#ifdef APSE_DEBUGGING
+            int k = APSE_IDX(i, ap->bitvectors_in_state, j);
+	    int l = ap->bytes_in_all_states/sizeof(apse_vec_t);
+	    assert (k < l);
+#endif
 	    APSE_BIT_SET(ap->prev_state, i, ap->bitvectors_in_state, j);
+	}
     }
 }
 
@@ -418,7 +431,9 @@ void apse_reset(apse_t *ap) {
     _apse_reset_state(ap);
 
     ap->text_position = ap->text_initial_position;
-    ap->text_position_range = APSE_MATCH_BAD;
+#if 0
+    ap->text_position_range = APSE_MATCH_BAD; /* Do not reset this. */
+#endif
 
     ap->match_state = APSE_MATCH_STATE_BOT;
     ap->match_begin = APSE_MATCH_BAD;
@@ -433,6 +448,9 @@ apse_bool_t apse_set_edit_distance(apse_t *ap, apse_size_t edit_distance) {
     if (ap->prev_state)
 	free(ap->prev_state);
 
+    if (edit_distance >= ap->pattern_size)
+        edit_distance = ap->pattern_size;
+
     ap->edit_distance = edit_distance;
 
     ap->bytes_in_all_states = (edit_distance + 1) * ap->bytes_in_state;
@@ -440,11 +458,11 @@ apse_bool_t apse_set_edit_distance(apse_t *ap, apse_size_t edit_distance) {
     ap->state = ap->prev_state = 0;
 
     ap->state = calloc(edit_distance + 1, ap->bytes_in_state);
-    if (!ap->state)
+    if (ap->state == 0)
 	goto out;
 
     ap->prev_state = calloc(edit_distance + 1, ap->bytes_in_state);
-    if (!ap->prev_state)
+    if (ap->prev_state == 0)
 	goto out;
 
     apse_reset(ap);
@@ -457,6 +475,8 @@ apse_bool_t apse_set_edit_distance(apse_t *ap, apse_size_t edit_distance) {
 
     if (ap->edit_distance && ap->bitvectors_in_state)
 	ap->largest_distance = ap->edit_distance * ap->bitvectors_in_state;
+    else
+	ap->largest_distance =  0;
 
     ap->match_begin_bitvector	=
 	(edit_distance + 1) / APSE_BITS_IN_BITVEC;
@@ -467,6 +487,17 @@ apse_bool_t apse_set_edit_distance(apse_t *ap, apse_size_t edit_distance) {
     ap->match_end_bitvector =
 	(ap->pattern_size - 1) / APSE_BITS_IN_BITVEC;
 
+#ifdef APSE_DEBUGGING
+    if (ap->has_different_distances) {
+	printf("(edit distances: ");
+	printf("insertions = %ld, deletions = %ld, substitutions = %ld)\n",
+	       ap->edit_insertions,
+	       ap->edit_deletions,
+	       ap->edit_substitutions);
+    } else
+	printf("(edit_distance = %ld)\n", ap->edit_distance);
+#endif
+
 out:
     return ap->state && ap->prev_state;
 }
@@ -488,13 +519,14 @@ apse_bool_t apse_set_exact_slice(apse_t*	ap,
 				 apse_ssize_t	exact_begin,
 				 apse_ssize_t	exact_size,
 				 apse_bool_t	exact) {
-    apse_ssize_t	i, j, true_begin, true_size;
+    apse_ssize_t	true_begin, true_size;
     apse_bool_t	okay = 0;
+    apse_size_t i, j;
 
     if (!ap->exact_mask) {
 
 	ap->exact_mask = calloc((size_t)1, ap->bytes_in_state);
-	if (!ap->exact_mask)
+	if (ap->exact_mask == 0)
 	    goto out;
 
 	ap->exact_positions = 0;
@@ -539,7 +571,7 @@ apse_bool_t apse_set_caseignore_slice(apse_t*		ap,
 
 	ap->fold_mask = calloc((apse_size_t)APSE_CHAR_MAX,
 			       ap->bytes_in_state);
-	if (!ap->fold_mask)
+	if (ap->fold_mask == 0)
 	    goto out;
 
 	memcpy(ap->fold_mask,
@@ -619,7 +651,7 @@ apse_t *apse_create(unsigned char*	pattern,
 	       pattern, pattern_size));
 
     ap = calloc((size_t)1, sizeof(*ap));
-    if (!ap)
+    if (ap == 0)
 	return 0;
 
     ap->pattern_size		= 0;
@@ -697,14 +729,7 @@ apse_t *apse_create(unsigned char*	pattern,
 	printf("(single bitvector");
     else
 	printf("(multiple bitvectors");
-    if (ap->has_different_distances) {
-	printf(")\n(edit distances: ");
-	printf("insertions = %ld, deletions = %ld, substitutions = %ld)\n",
-	       ap->edit_insertions,
-	       ap->edit_deletions,
-	       ap->edit_substitutions);
-    } else
-	printf(", edit_distance = %ld)\n", ap->edit_distance);
+    printf(")\n");
 #endif
     okay = 1;
 
@@ -883,7 +908,7 @@ static apse_bool_t _apse_match_next_state(apse_t *ap) {
 	    for (h = 0;
 		 h <= k;
 		 h += ap->bitvectors_in_state) {
-		for (i = h, j = i + ap->bitvectors_in_state - 1; i < j; j--)
+	        for (i = h, j = h + ap->bitvectors_in_state - 1; i < j; j--)
 		    if (ap->state[j] != ap->prev_state[j])
 			break;
 		if (ap->prev_state[j] == ap->state[j])
@@ -891,11 +916,28 @@ static apse_bool_t _apse_match_next_state(apse_t *ap) {
 		if (ap->state[j])
 		    active++;
 	    }
-	    if (equal == ap->edit_distance + 1 && ap->is_greedy == 0 ||
-		equal == 0 && ap->prev_active && active > ap->prev_active)
+#ifdef APSE_DEBUGGING
+	    printf("(equal = %d, active = %d)\n", equal, active);
+#endif
+	    if ((equal == ap->edit_distance + 1 &&
+		 ap->is_greedy == 0)
+		||
+		(equal < ap->prev_equal &&
+		 ap->prev_active &&
+		 active > ap->prev_active &&
+		 ap->text_position - ap->match_begin < 8 * ap->bytes_in_state &&
+		 !APSE_BIT_TST(ap->state,
+			       ap->edit_distance,
+			       ap->bitvectors_in_state,
+			       ap->text_position - ap->match_begin))) {
 		ap->match_begin = ap->text_position;
+#ifdef APSE_DEBUGGING
+		printf("(slide begin %d)\n", ap->match_begin);
+#endif
+	    }
 	    else if (active == 0)
 		_apse_match_fail(ap);
+	    ap->prev_equal  = equal;
 	    ap->prev_active = active;
 	}
 	break;
@@ -926,7 +968,7 @@ static apse_bool_t _apse_match_next_state(apse_t *ap) {
 	ap->match_state = APSE_MATCH_STATE_END;
 	APSE_DEBUG(printf("(match state %s)\n", apse_match_state_name(ap)));
 	ap->match_end	= ap->text_position - 1;
-	APSE_DEBUG(printf("(match end %ld)\n", ap->match_end));
+	APSE_DEBUG(printf("(match end %ld)\n",  ap->match_end));
     }
 
     return ap->match_state;
@@ -1314,8 +1356,8 @@ static apse_bool_t _apse_match(apse_t 		*ap,
 		    break;
 		previous_edit_distance = next_edit_distance;
 	    } 
-	    minimal_edit_distance = 0;
-	    if (next_edit_distance) {
+	    minimal_edit_distance = next_edit_distance;
+	    if (next_edit_distance > 1) {
 		do {
 		    minimal_edit_distance =
 			(previous_edit_distance + next_edit_distance) / 2;
@@ -1362,7 +1404,6 @@ apse_bool_t apse_match_next(apse_t *ap,
 apse_ssize_t apse_index(apse_t *ap,
 			 unsigned char *text, apse_size_t text_size) {
     apse_size_t did_match = _apse_match(ap, text, text_size);
-
     _apse_match_eot(ap);
     ap->match_state = APSE_MATCH_STATE_BOT;
 
